=== modified file 'shelf.py'
--- shelf.py	2005-06-16 14:44:53 +0000
+++ shelf.py	2005-08-17 15:47:08 +0000
@@ -1,40 +1,32 @@
 #!/usr/bin/python
 
-# Copyright (C) 2005 Michael Ellerman <michael@ellerman.id.au>
-#
-#    This program is free software; you can redistribute it and/or modify
-#    it under the terms of the GNU General Public License as published by
-#    the Free Software Foundation; either version 2 of the License, or
-#    (at your option) any later version.
-#
-#    This program is distributed in the hope that it will be useful,
-#    but WITHOUT ANY WARRANTY; without even the implied warranty of
-#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-#    GNU General Public License for more details.
-#
-#    You should have received a copy of the GNU General Public License
-#    along with this program; if not, write to the Free Software
-#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-import patches
+from patches import parse_patches
 import os
 import sys
 import string
+import tty, termios
 from bzrlib.commands import Command
 
-def parse_args(args):
-    if len(args) == 2 and args[1] == '--bzr-usage':
-        print '\n'
-        return True
-    elif len(args) == 2 and args[1] == '--bzr-help':
-        print 'Shelve a patch, you can get it back later with unshelve.'
-        return True
-    elif len(args) == 1:
-        pass
-    else:
-        raise Exception("Don't understand args %s" % args)
-
-    return False
+def main(args):
+    name = os.path.basename(args.pop(0))
+
+    if name not in ['shelve', 'unshelve']:
+        raise Exception("Unknown command name '%s'" % name)
+
+    if len(args) > 0:
+        if args[0] == '--bzr-usage':
+            print '\n'
+            return 0
+        elif args[0] == '--bzr-help':
+            print 'Shelve a patch, you can get it back later with unshelve.'
+            return 0
+        else:
+            raise Exception("Don't understand args %s" % args)
+
+    if eval(name + "()"):
+        return 0
+
+    return 1
 
 def unshelve():
     root = run_bzr('root')[0].strip()
@@ -59,12 +51,17 @@
 
     return True
 
+class QuitException(Exception):
+    pass
+
 def shelve():
-    diff_lines = run_bzr('diff')
-    patch_list = patches.parse_patches(diff_lines.__iter__())
-    to_shelve = prompt_user(patch_list)
+    patches = parse_patches(run_bzr('diff'))
+    try:
+        patches = HunkSelector(patches).select()
+    except QuitException:
+        return False
 
-    if len(to_shelve) == 0:
+    if len(patches) == 0:
         print >>sys.stderr, 'Nothing to shelve'
         return True
 
@@ -73,7 +70,7 @@
     print >>sys.stderr, "Saving shelved patches to", shelf
     shelf = open(shelf, 'a')
 
-    for patch in to_shelve:
+    for patch in patches:
         shelf.write(str(patch))
 
     shelf.flush()
@@ -82,7 +79,7 @@
 
     print >>sys.stderr, "Reverting shelved patches"
     pipe = os.popen('patch -d %s -sR -p0' % root, 'w')
-    for patch in to_shelve:
+    for patch in patches:
        pipe.write(str(patch))
     pipe.flush()
 
@@ -103,42 +100,6 @@
         raise Exception("Failed running bzr")
     return lines
 
-def prompt_user(patch_list):
-    total = 0
-    for patch in patch_list:
-        total += len(patch.hunks)
-
-    out = sys.stdout
-    inp = sys.stdin
-
-    to_shelve = []
-    i = 1
-    for patch in patch_list:
-        if patch.oldname != patch.newname:
-            name = "%s -> %s" % (patch.oldname, patch.newname)
-        else:
-            name = patch.oldname
-
-        hunks = patch.hunks[:]
-        for hunk in hunks:
-            print name
-            print hunk
-            while True:
-                out.write('Shelve this change? (%d of %d) [yn] ' % (i, total))
-                line = inp.readline().strip().lower()
-                if line == 'y' or line == 'yes':
-                    break
-                elif line == 'n' or line == 'no':
-                    patch.hunks.remove(hunk)
-                    break
-
-            i += 1
-
-        if len(patch.hunks) > 0:
-            to_shelve.append(patch)
-
-    return to_shelve
-
 class cmd_shelve(Command):
     """Temporarily remove some changes from the current tree.
     Use 'unshelve' to restore these changes.
@@ -153,3 +114,189 @@
     """
     def run(self):
         return unshelve()
+
+class HunkSelector:
+    class Option:
+        def __init__(self, char, action, help, default=False):
+            self.char = char
+            self.action = action
+            self.default = default
+            self.help = help
+
+    standard_options = [
+        Option('n', 'shelve', 'shelve this change for the moment.',
+            default=True),
+        Option('y', 'keep', 'keep this change in your tree.'),
+        Option('d', 'done', 'done, skip to the end.'),
+        Option('i', 'invert', 'invert the current selection of all hunks.'),
+        Option('s', 'status', 'show status of hunks.'),
+        Option('q', 'quit', 'quit')
+    ]
+
+    end_options = [
+        Option('y', 'continue', 'proceed to shelve selected changes.',
+            default=True),
+        Option('r', 'restart', 'restart the hunk selection loop.'),
+        Option('s', 'status', 'show status of hunks.'),
+        Option('i', 'invert', 'invert the current selection of all hunks.'),
+        Option('q', 'quit', 'quit')
+    ]
+
+    def __init__(self, patches):
+        self.patches = patches
+        self.total_hunks = 0
+        for patch in patches:
+            for hunk in patch.hunks:
+                # everything's shelved by default
+                hunk.selected = True
+                self.total_hunks += 1
+
+    def __get_option(self, char):
+        for opt in self.standard_options:
+            if opt.char == char:
+                return opt
+        raise Exception('Option "%s" not found!' % char)
+
+    def __select_loop(self):
+        j = 0
+        for patch in self.patches:
+            i = 0
+            lasti = -1
+            while i < len(patch.hunks):
+                hunk = patch.hunks[i]
+                if lasti != i:
+                    print patch.get_header(), hunk
+                    j += 1
+                lasti = i
+
+                prompt = 'Keep this change? (%d of %d)' \
+                            % (j, self.total_hunks)
+
+                if hunk.selected:
+                    self.__get_option('n').default = True
+                    self.__get_option('y').default = False
+                else:
+                    self.__get_option('n').default = False
+                    self.__get_option('y').default = True
+
+                action = self.__ask_user(prompt, self.standard_options)
+
+                if action == 'keep':
+                    hunk.selected = False
+                elif action == 'shelve':
+                    hunk.selected = True
+                elif action == 'done':
+                    return True
+                elif action == 'invert':
+                    self.__invert_selection()
+                    self.__show_status()
+                    continue
+                elif action == 'status':
+                    self.__show_status()
+                    continue
+                elif action == 'quit':
+                    return False
+
+                i += 1
+        return True
+
+    def select(self):
+        if self.total_hunks == 0:
+            return []
+
+        done = False
+        while not done:
+            if not self.__select_loop():
+                return []
+
+            while True:
+                self.__show_status()
+                prompt = "Shelve these changes, or restart?"
+                action = self.__ask_user(prompt, self.end_options)
+
+                if action == 'continue':
+                    done = True
+                    break
+                elif action == 'quit':
+                    return []
+                elif action == 'status':
+                    self.__show_status()
+                elif action == 'invert':
+                    self.__invert_selection()
+                elif action == 'restart':
+                    break
+
+
+        for patch in self.patches:
+            tmp = []
+            for hunk in patch.hunks:
+                if hunk.selected:
+                    tmp.append(hunk)
+            patch.hunks = tmp
+
+        tmp = []
+        for patch in self.patches:
+            if len(patch.hunks):
+                tmp.append(patch)
+        self.patches = tmp
+
+        return self.patches
+
+    def __invert_selection(self):
+        for patch in self.patches:
+            for hunk in patch.hunks:
+                if hunk.__dict__.has_key('selected'):
+                    hunk.selected = not hunk.selected
+                else:
+                    hunk.selected = True
+
+    def __show_status(self):
+        print '\nStatus:'
+        for patch in self.patches:
+            print '  %s' % patch.oldname
+            shelve = 0
+            keep = 0
+            for hunk in patch.hunks:
+                if hunk.selected:
+                    shelve += 1
+                else:
+                    keep += 1
+
+            print '    %d hunks to be shelved' % shelve
+            print '    %d hunks to be kept' % keep
+            print
+
+    def __getchar(self):
+        fd = sys.stdin.fileno()
+        settings = termios.tcgetattr(fd)
+        try:
+            tty.setraw(fd)
+            ch = sys.stdin.read(1)
+        finally:
+            termios.tcsetattr(fd, termios.TCSADRAIN, settings)
+        return ch
+
+    def __ask_user(self, prompt, options):
+        while True:
+            sys.stdout.write(prompt)
+            sys.stdout.write(' [')
+            for opt in options:
+                if opt.default:
+                    default = opt
+                sys.stdout.write(opt.char)
+            sys.stdout.write('?] (%s): ' % default.char)
+
+            response = self.__getchar()
+
+            # default, which we see as newline, is 'n'
+            if response in ['\n', '\r', '\r\n']:
+                response = default.char
+
+            print response # because echo is off
+
+            for opt in options:
+                if opt.char == response:
+                    return opt.action
+
+            for opt in options:
+                print '  %s - %s' % (opt.char, opt.help)

=== modified file 'shelve'
--- shelve	2005-06-15 15:25:13 +0000
+++ shelve	2005-08-17 15:47:08 +0000
@@ -3,13 +3,5 @@
 import sys
 import shelf
 
-def main(args):
-    finished = shelf.parse_args(args)
-
-    if finished:
-        return True
-    else:
-        return shelf.shelve()
-
 if __name__ == '__main__':
-    sys.exit(main(sys.argv) and 0 or 1)
+    sys.exit(shelf.main(sys.argv))

=== modified file 'unshelve'
--- unshelve	2005-06-15 15:25:13 +0000
+++ unshelve	2005-08-17 15:47:08 +0000
@@ -3,13 +3,5 @@
 import sys
 import shelf
 
-def main(args):
-    finished = shelf.parse_args(args)
-
-    if finished:
-        return True
-    else:
-        return shelf.unshelve()
-
 if __name__ == '__main__':
-    sys.exit(main(sys.argv) and 0 or 1)
+    sys.exit(shelf.main(sys.argv))

